import json
from threading import Thread, Event
import time


class KMemoryContentReceiverThread(Thread):

    def __init__(self, memory=None, consumer=None, topic_config=None):
        super(KMemoryContentReceiverThread, self).__init__()
        self.memory = memory
        self.consumer = consumer
        self.topic_config = topic_config
        self._stop_event = Event()
        self.stopped = False

    def stop(self):
        self.stopped = True
        self._stop_event.set()

    def is_stop(self):
        self._stop_event.is_set()

    def run(self):

        while not self.stopped:
            message = self.consumer.poll(10)

            if message is None:
                continue
            if message.error():
                print('Consumer error: %s' % message.error())
                continue

            j = json.loads(message.value().decode('utf-8'))

            self.memory.set_evaluation(j["evaluation"])
            self.memory.set_i(j["i"])

            time.sleep(0.01)



from pyctm.memory.kafka.k_distributed_memory_behavior import KDistributedMemoryBehavior


class TopicConfig:

    def __init__(self, broker, name, k_distributed_memory_behavior=KDistributedMemoryBehavior.PULLED,
                 regex_pattern=None, class_name=None) -> None:
        self.name = name
        self.k_distributed_memory_behavior = k_distributed_memory_behavior
        self.class_name = class_name
        self.broker = broker
        self.regex_pattern = regex_pattern



from confluent_kafka import Producer
from confluent_kafka.admin import AdminClient
from confluent_kafka.cimpl import NewTopic


class KProducerBuilder:

    @staticmethod
    def build_producer(broker):
        producer = Producer({
            'bootstrap.servers': broker
        })
        return producer

    @staticmethod
    def check_topic_exist(broker, topic):
        kafka_admin = AdminClient({"bootstrap.servers": broker})
        topic_metadata = kafka_admin.list_topics()
        if topic_metadata.topics.get(topic) is None:
            new_kafka_topic = NewTopic(topic, num_partitions=1, replication_factor=1)
            kafka_admin.create_topics([new_kafka_topic])

    @staticmethod
    def generate_producers(topic_configs):
        producers = []

        for topic_config in topic_configs:
            print('Creating producer for topic configuration - Name: %s - Broker: %s - Class: %s - Behavior Type: %s' %
                  (topic_config.name,
                   topic_config.broker,
                   topic_config.class_name,
                   topic_config.k_distributed_memory_behavior.name))

            producer = KProducerBuilder.build_producer(topic_config.broker)

            print('Producer created fo topic %s.' % topic_config.name)

            producers.append(producer)

        return producers



import threading

from pyctm.memory.distributed_memory_type import DistributedMemoryType
from pyctm.memory.kafka.builder.k_producer_builder import KProducerBuilder
from pyctm.memory.kafka.thread.k_memory_content_publisher_thread import KMemoryContentPublisherThread
from pyctm.memory.memory import Memory
from pyctm.memory.kafka.builder.k_consumer_builder import KConsumerBuilder
from pyctm.memory.memory_object import MemoryObject
from pyctm.memory.kafka.thread.k_memory_content_receiver_thread import KMemoryContentReceiverThread


class KDistributedMemory(Memory):

    def __init__(self, name="", distributed_memory_type=DistributedMemoryType.INPUT_MEMORY, topics_config=[]):
        self.name = name
        self.distributed_memory_type = distributed_memory_type
        self.topics_config = topics_config

        print('Creating KDistributedMemory %s for type %s.' %
              (name, distributed_memory_type))

        self.memories = []
        self.k_memory_content_receiver_threads = []
        self.k_memory_content_publisher_threads = []

        self.condition = threading.Condition()

        self.init_memory()

        print('KDistributeMemory %s created.' % name)

    def init_memory(self):

        if self.distributed_memory_type == DistributedMemoryType.INPUT_MEMORY \
                or self.distributed_memory_type == DistributedMemoryType.INPUT_BROADCAST_MEMORY:
            self.consumers_setup(self.topics_config)
        else:
            self.producers_setup(self.topics_config)

    def consumers_setup(self, topics_config):
        print('Creating the consumers.')

        topics_consumer_map = KConsumerBuilder.generate_consumers(
            topics_config, self.name)

        for topic_config, consumer in topics_consumer_map.items():
            memory = MemoryObject(0, topic_config.name)
            self.memories.append(memory)

            k_memory_content_receiver_thread = KMemoryContentReceiverThread(
                memory, consumer, topics_config)
            k_memory_content_receiver_thread.start()

            self.k_memory_content_receiver_threads.append(
                k_memory_content_receiver_thread)

        print('Consumers created.')

    def producers_setup(self, topics_config):
        print('Creating the producers.')

        producers = KProducerBuilder.generate_producers(topics_config)

        for (producer, topic_config) in zip(producers, topics_config):
            memory = MemoryObject(0, topic_config.name)
            self.memories.append(memory)

            k_memory_content_publisher_thread = KMemoryContentPublisherThread(memory, producer, topic_config,
                                                                              self.condition)
            k_memory_content_publisher_thread.start()

            self.k_memory_content_publisher_threads.append(k_memory_content_publisher_thread)

        print('Producers created.')

    def stop(self):
        for k_thread in self.k_memory_content_receiver_threads:
            k_thread.stop()

        for k_thread in self.k_memory_content_publisher_threads:
            k_thread.stop()


    def get_i(self):
        memory = max(self.memories, key=lambda m: m.get_i())
        return memory.get_evaluation()

    def get_i_index(self, index):
        try:
            return self.memories[index]
        except IndexError:
            print('Impossible to get memory content. Index %s out of bounds.' % index)
            return None

    def get_memory(self):
        return max(self.memories, key=lambda memory: memory.get_i())

    def set_i(self, i):
        try:
            self.condition.acquire()
            self.memories[0].set_i(i)
            self.condition.notify()
        except IndexError:
            print('Impossible to set memory content. Index 0 out of bounds.')
        finally:
            self.condition.release()

    def set_i_index(self, i, index):
        try:
            self.condition.acquire()
            self.memories[index].set_i(i)
            self.condition.notify()
        except IndexError:
            print('Impossible to get memory content. Index %s out of bounds.' % index)
        finally:
            self.condition.release()

    def set_i_evaluation_index(self, i, evaluation, index):
        try:
            self.condition.acquire()
            self.memories[index].set_i(i)
            self.memories[index].set_evaluation(evaluation)
            self.condition.notify()
        except IndexError:
            print('Impossible to get memory content. Index %s out of bounds.' % index)
        finally:
            self.condition.release()

    def get_evaluation(self):
        memory = max(self.memories, key=lambda m: m.get_i())
        if memory is not None:
            return memory.get_evaluation()

        return -1

    def set_evaluation(self, evaluation):
        try:
            self.condition.acquire()
            self.memories[0].set_evaluation(evaluation)
            self.condition.notify()
        except IndexError:
            print('Impossible to get memory content. Index 0 out of bounds.')
        finally:
            self.condition.release()

    def set_evaluation_index(self, evaluation, index):
        try:
            self.condition.acquire()
            self.memories[index].set_evaluation(evaluation)
            self.condition.notify()
        except IndexError:
            print('Impossible to get memory content. Index %s out of bounds.' % index)
        finally:
            self.condition.release()

    def get_name(self):
        return self.name

    def set_name(self, name):
        self.name = name



from pyctm.memory.distributed_memory_type import DistributedMemoryType
from pyctm.memory.k_distributed_memory import KDistributedMemory


class CodeletApplication():

    def __init__(self, codelet_memories):
        self.codelet_memories = codelet_memories
        self.codelets = []

        self.initialize_distributed_memories()

    def initialize_distributed_memories(self):

        print('Inserting codelets into Codelet Application.')

        for codelet, memories in self.codelet_memories.items():
            for memory in memories:

                if isinstance(memory, KDistributedMemory):

                    if memory.distributed_memory_type == DistributedMemoryType.INPUT_MEMORY:
                        if len(list(filter(lambda input_memory: (input_memory == memory), codelet.inputs))) > 0:
                            codelet.inputs.append(memory)

                    elif memory.distributed_memory_type == DistributedMemoryType.OUTPUT_MEMORY:
                        if len(list(filter(lambda output_memory: (output_memory == memory), codelet.outputs))) > 0:
                            codelet.outputs.append(memory)
                    else:
                        if len(list(filter(lambda broadcast_memory: (broadcast_memory == memory), codelet.broadcast))) > 0:
                            codelet.broadcast.append(memory)

            self.codelets.append(codelet)

            print('Starting codelet %s' % codelet.name)
            codelet.start()
            print('Codelet %s started.' % codelet.name)





import re
from confluent_kafka.admin import AdminClient
from pyctm.memory.kafka.config.topic_config import TopicConfig
from pyctm.memory.kafka.k_distributed_memory_behavior import KDistributedMemoryBehavior


class TopicConfigProvider:

    @staticmethod
    def generate_topic_configs_regex_pattern(broker, regex_pattern, class_name):

        k_admin = AdminClient({
            'bootstrap.servers': broker
        })

        cluster_metadata = k_admin.list_groups().topics

        if len(cluster_metadata.items()) == 0:
            raise Exception('Topics not found. Review regex pattern.')

        found_topics = []

        for key, _ in cluster_metadata.items():
            if re.search(regex_pattern, key):
                found_topics.append(TopicConfig(
                    key, broker, KDistributedMemoryBehavior.PULLED, regex_pattern, class_name))

        return found_topics



class Memory:

    def get_i(self) -> object:
        pass

    def get_name(self) -> str:
        pass

    def get_evaluation(self) -> float:
        pass

    def get_id(self) -> str:
        pass

    def set_i(self, i) -> None:
        pass

    def set_name(self, name) -> None:
        pass

    def set_evaluation(self, evaluation) -> None:
        pass

    def set_id(self, id) -> None:
        pass



from threading import Thread

import time


class Codelet(Thread):
    def __init__(self, name):
        Thread.__init__()
        self.activation = 0
        self.threshold = 0
        self.name = name
        self.enabled = True
        self.timestamp = 300

        self.inputs = []
        self.outputs = []
        self.broadcast = []

    def proc(self):
        pass

    def run(self):

        while self.enabled:
            self.proc()
            time.sleep(self.timestamp)

    def add_input(self, input):
        self.inputs.append(input)

    def add_output(self, output):
        self.outputs.append(output)

    def get_activation(self):
        return self.activation

    def set_activation(self, activation):

        if activation > 1:
            self.activation = 1
        elif activation < 0:
            self.activation = 0
        else:
            self.activation = activation

    def get_inputs(self):
        return self.inputs

    def get_outputs(self):
        return self.outputs

    def get_broadcast(self):
        return self.broadcast



from setuptools import find_packages, setup
setup(
    name='pyctm',
    packages=find_packages(),
    version='0.0.4',
    description='Python Cognitive System Toolkit for Microservices',
    author='Eduardo de Moraes Froes',
    license='MIT',
    install_requires=['confluent-kafka'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest==4.4.1'],
    test_suite='tests',
    keywords=['python', 'cst']
)



import json
from threading import Thread, Event
import time

from pyctm.memory.kafka.builder.k_producer_builder import KProducerBuilder
from pyctm.memory.kafka.k_distributed_memory_behavior import KDistributedMemoryBehavior


class KMemoryContentPublisherThread(Thread):

    def __init__(self, memory=None, producer=None, topic_config=None, condition=None):
        super(KMemoryContentPublisherThread, self).__init__()
        self.memory = memory
        self.producer = producer
        self.topic_config = topic_config
        self.last_i = None
        self.last_evaluation = 0
        self.condition = condition
        self._stop_event = Event()
        self.stopped = False

        KProducerBuilder.check_topic_exist(topic_config.broker, topic_config.name)

    def stop(self):
        self.stopped = True
        self.condition.acquire()
        self.condition.notify()
        self._stop_event.set()

    def is_stop(self):
        return self._stop_event.is_set()

    def run(self):

        print('Content publisher thread initialized for memory %s.' % self.memory.get_name())

        while not self.stopped:
            if self.topic_config.k_distributed_memory_behavior == KDistributedMemoryBehavior.TRIGGERED:
                self.condition.acquire()
                self.condition.wait()

                object_json = json.dumps(self.memory.__dict__)

                self.producer.poll(10)
                self.producer.produce(self.topic_config.name, object_json)

                self.condition.release()
            else:

                if self.memory.get_i() != self.last_i or self.memory.get_evaluation() != self.last_evaluation:
                    object_json = json.dumps(self.memory.__dict__)

                    self.producer.poll(10)
                    self.producer.produce(self.topic_config.name, object_json)

                    self.last_evaluation = self.memory.get_evaluation()
                    self.last_i = self.memory.get_i()

            time.sleep(0.01)




from confluent_kafka import Consumer
from confluent_kafka.admin import AdminClient
from confluent_kafka.cimpl import NewTopic

from pyctm.memory.kafka.topic_config_provider import TopicConfigProvider


class KConsumerBuilder:

    @staticmethod
    def build_consumer(broker, consumer_group_id, topic_config):

        KConsumerBuilder.check_topic_exist(broker, topic_config.name)

        consumer = Consumer({
            'bootstrap.servers': broker,
            'group.id': consumer_group_id,
            'auto.offset.reset': 'earliest'
        })

        consumer.subscribe([topic_config.name])

        return consumer

    @staticmethod
    def check_topic_exist(broker, topic):
        kafka_admin = AdminClient({"bootstrap.servers": broker})
        topic_metadata = kafka_admin.list_topics()
        if topic_metadata.topics.get(topic) is None:
            new_kafka_topic = NewTopic(topic, num_partitions=1, replication_factor=1)
            kafka_admin.create_topics([new_kafka_topic])

    @staticmethod
    def generate_consumers(topic_configs, consumer_group_id):

        consumers = {}

        for topic_config in topic_configs:

            if topic_config.regex_pattern is not None:
                print('Regex pattern %s identified.' %
                      topic_config.regex_pattern)

                if not topic_config.regex_pattern:

                    found_topic_configs = TopicConfigProvider.generate_topic_configs_regex_pattern(
                        topic_config.broker, topic_config.regex_pattern, topic_config.class_name)

                    if len(found_topic_configs) == 0:
                        raise Exception(
                            'Topic regex not found - pattern - %s' % topic_config.regex_pattern)

                    regex_pattern_consumers = KConsumerBuilder.generate_consumers(
                        found_topic_configs, consumer_group_id)

                    for key, value in regex_pattern_consumers.items():
                        consumers[key] = value

                return

            print('Creating consumer for topic configuration - Name: %s - Broker: %s - Class: %s - Behavior Type: %s'
                  % (topic_config.name, topic_config.broker, topic_config.class_name,
                     topic_config.k_distributed_memory_behavior))

            consumer = KConsumerBuilder.build_consumer(
                topic_config.broker, consumer_group_id, topic_config)

            print('Consumer created for topic %s.' % topic_config.name)

            consumers[topic_config] = consumer

        return consumers



import unittest
import time
from pyctm.memory.distributed_memory_type import DistributedMemoryType


from pyctm.memory.k_distributed_memory import KDistributedMemory
from pyctm.memory.kafka.config.topic_config import TopicConfig
from pyctm.memory.kafka.k_distributed_memory_behavior import KDistributedMemoryBehavior


class KDistributedMemoryTest(unittest.TestCase):

    @staticmethod
    def memory_initialization():

        input_topic_configs = []
        output_topic_configs = []

        input_topic_configs.append(TopicConfig("localhost:9092", "topic-1", KDistributedMemoryBehavior.PULLED))
        output_topic_configs.append(TopicConfig("localhost:9092", "topic-1", KDistributedMemoryBehavior.TRIGGERED))

        distributed_output_memory = KDistributedMemory("MEMORY_TEST_OUTPUT", DistributedMemoryType.OUTPUT_MEMORY,
                                                       output_topic_configs)
        distributed_input_memory = KDistributedMemory("MEMORY_TEST_INPUT", DistributedMemoryType.INPUT_MEMORY,
                                                      input_topic_configs)

        return distributed_input_memory, distributed_output_memory

    def test_should_has_message_in_input_distributed_memory(self):
        distributed_input_memory, distributed_output_memory = self.memory_initialization()

        message = "Test message in the distributed memory!"

        distributed_output_memory.set_i(message)

        time.sleep(2)
        assert distributed_output_memory.get_i() == distributed_input_memory.get_i()

        message = "New message to test!"
        distributed_output_memory.set_i(message)

        time.sleep(2)
        assert distributed_output_memory.get_i() == distributed_input_memory.get_i()

        distributed_output_memory.stop()
        distributed_input_memory.stop()

if __name__ == '__main__':
    unittest.main()


import time
import unittest
from pyctm.memory import memory_object
from pyctm.memory.distributed_memory_type import DistributedMemoryType
from pyctm.memory.k_distributed_memory import KDistributedMemory
from pyctm.memory.kafka.config.topic_config import TopicConfig
from pyctm.memory.kafka.k_distributed_memory_behavior import KDistributedMemoryBehavior

# from pyctm.memory.memory import MemoryObject
from pyctm.memory.memory_object import MemoryObject


class MemoryObjectTest(unittest.TestCase):

    def test_memory_initialization(self):
        memoryObject = MemoryObject(1, "test", 10, 1)

        assert memoryObject.get_id() == 1
        assert memoryObject.get_name() == "test"
        assert memoryObject.get_i() == 10
        assert memoryObject.get_evaluation() == 1

    def test_memory_set_evaluation(self):
        memoryObject = MemoryObject(1, "test", 10, 10)
        assert memoryObject.get_evaluation() == 1

        memoryObject.set_evaluation(-10)
        assert memoryObject.get_evaluation() == 0


if __name__ == '__main__':
    unittest.main()



from pyctm.memory.memory import Memory


class MemoryObject(Memory):

    def __init__(self, id=0, name="", i=None, evaluation=0):
        self.id = id
        self.name = name
        self.i = i
        self.locked = False
        self.set_evaluation(evaluation)

    def get_i(self):
        return self.i

    def get_name(self):
        return self.name

    def get_evaluation(self):
        return self.evaluation

    def get_id(self):
        return self.id

    def set_i(self, i):
        self.i = i

        if self.locked:
            self.locked = False

    def set_name(self, name):
        self.name = name

    def set_evaluation(self, evaluation):
        if evaluation > 1:
            self.evaluation = 1
        elif evaluation < 0:
            self.evaluation = 0
        else:
            self.evaluation = evaluation

    def set_id(self, id):
        self.id = id
